@@ -2,6 +2,7 @@ from IPython.display import display_html
 from math import sqrt
 from numpy import triu, ones, nan, inf
 from pandas import DataFrame, Series, notna, crosstab
+from random import shuffle
 from scipy.stats import kruskal, chi2_contingency
 from sklearn.base import BaseEstimator, TransformerMixin
 from statsmodels.formula.api import ols
@@ -21,41 +22,67 @@ class FeatureSelector(BaseEstimator, TransformerMixin):
         Features on which to compute association.
     n_best, int:
         Number of features to be selected
+    sample_size: float, default 1.
+        Should be set between ]0, 1]
+        Size of sampled list of features speeds up computation. 
+        By default, all features are used. For sample_size=0.5,
+        FeatureSelector will search for the best features in 
+        features[:len(features)//2] and then in features[len(features)//2:]
     measures, List[Callable]: default list().
         List of association measures to be used.
         Implemented measures are:
-            - For association evaluation: `kruskal_measure`, `R_measure`
-            - For outlier detection: `zscore_measure`, `iqr_measure`
+            [Quantitative Features] 
+             - For association evaluation: `kruskal_measure`, `R_measure`
+             - For outlier detection: `zscore_measure`, `iqr_measure`
+            [Qualitative Features] 
+             - For correlation: `chi2_measure`, `cramerv_measure`, `tschuprowt_measure`
         Ranks features based on last measure of the list.
     filters, List[Callable]: default list().
         List of filters to be used.
         Implemented filters are:
-          - For linear correlation: `spearman_filter`, `pearson_filter`
-          - For multicoloinearity: `vif_filter`
+            [Quantitative Features] 
+             - For linear correlation: `spearman_filter`, `pearson_filter`
+             - For multicoloinearity: `vif_filter`
+            [Qualitative Features] 
+             - For correlation: `cramerv_filter`, `tschuprowt_filter`
 
-    params examples
-    ---------------
+    Thresholds (to be passed as kwargs)
+    ----------
+    thresh_measure, float: default 0.
+        Minimum association between target and features
+        To be used with: `measure_filter`
+    name_measure, str
+        Measure to be used for minimum association filtering
+        To be used with: `measure_filter`
     thresh_nan, float: default 1.
         Maximum percentage of NaNs in a feature
+        To be used with: `nans_measure`
     thresh_mode, float: default 1.
         Maximum percentage of the mode of a feature
+        To be used with: `mode_measure`
     thresh_outlier, float: default 1.
         Maximum percentage of Outliers in a feature
+        To be used with: `iqr_measure`, `zscore_measure`
     thresh_corr, float: default 1.
         Maximum association between features
+        To be used with: `spearman_filter`, `pearson_filter`, `cramerv_filter`, `tschuprowt_filter`
     thresh_vif, float: default inf
         Maximum VIF between features
-    ascending, bool: default False
-        Ascending of Descending sort by sort_measure
+        To be used with: `vif_filter`
+    ascending, bool default False
+        According to this measure:
+         - True: Lower values of the measure are to be considered as more associated to the target
+         - False: Higher values of the measure are to be considered as more associated to the target
     """
     
     def __init__(self, features: List[str], n_best: int, measures: List[Callable]=list(), filters: List[Callable]=list(),
-                 copy: bool=True, verbose: bool=True, **params) -> None:
+                 sample_size: float=1., copy: bool=True, verbose: bool=True, **params) -> None:
         
         self.features = features[:]
         self.n_best = n_best
         assert n_best <= len(features), "Must set n_best <= len(features)"
         self.best_features = features[:]
+        self.sample_size = sample_size
         
         self.measures = [dtype_measure, nans_measure, mode_measure] + measures[:]
         self.filters = [thresh_filter] + filters[:]
@@ -77,17 +104,19 @@ class FeatureSelector(BaseEstimator, TransformerMixin):
             
         return association
     
-    def measure_apply(self, X: DataFrame, y: Series) -> None:
+    def measure_apply(self, X: DataFrame, y: Series, features: List[str]) -> None:
         """ Measures association between columns of X and y
 
     Parameters
     ----------
-    ascending, bool: default False
-      Ascending of Descending sort by sort_measure
+    ascending, bool default False
+        According to this measure:
+         - True: Lower values of the measure are to be considered as more associated to the target
+         - False: Higher values of the measure are to be considered as more associated to the target
     """
         
         # applying association measure to each column
-        self.associations = X[self.features].apply(self.measure, y=y, result_type='expand', axis=0).T
+        self.associations = X[features].apply(self.measure, y=y, result_type='expand', axis=0).T
         
         # filtering non association measure (pct_zscore, pct_iqr...)
         asso_measures = [c for c in self.associations if '_measure' in c]
@@ -103,8 +132,10 @@ class FeatureSelector(BaseEstimator, TransformerMixin):
 
     Parameters
     ----------
-    ascending, bool: default False
-      Ascending of Descending sort by sort_measure
+    ascending, bool default False
+        According to this measure:
+         - True: Lower values of the measure are to be considered as more associated to the target
+         - False: Higher values of the measure are to be considered as more associated to the target
     """
         
         # ordering features by sort_by
@@ -126,12 +157,11 @@ class FeatureSelector(BaseEstimator, TransformerMixin):
         style = style.set_caption(caption)
         display_html(style._repr_html_(), raw=True)
         
-    
-    def fit(self, X: DataFrame, y: Series) -> None:
-        """ Selects the n_best features"""
+    def fit_features(self, X: DataFrame, y: Series, features: List[str], n_best: int) -> List[str]:
+        """ Selects the n_best features amongst the specified ones"""
         
         # initial computation of all association measures
-        self.measure_apply(X, y)
+        self.measure_apply(X, y, features)
 
         # displaying association measure
         if self.verbose:
@@ -150,10 +180,48 @@ class FeatureSelector(BaseEstimator, TransformerMixin):
             if n == 0 and self.verbose and len(self.filters) > 1:
                 self.display_stats(self.filtered_associations, 'Filtered association')
 
-        # retrieving the n_best features per ranking
+        # retrieving the n_best features per each ranking
+        best_features = []
         if len(self.sort_measures) > 0:
-            self.best_features = [feature for rank in ranks for feature in rank[:self.n_best]]
-            self.best_features = list(set(self.best_features))  # deduplicating
+            best_features = [feature for rank in ranks for feature in rank[:n_best]]
+            best_features = list(set(best_features))  # deduplicating
+        
+        return best_features
+    
+    def fit(self, X: DataFrame, y: Series) -> None:
+        """ Selects the n_best features"""
+        
+        # splitting features in chunks
+        if self.sample_size < 1:
+            
+            # shuffling features to get random samples of features
+            shuffle(self.features)
+
+            # number of features per sample
+            chunks = int(len(self.features) // (1 / self.sample_size))
+
+            # splitting feature list in samples
+            feature_samples = [self.features[chunks * i: chunks * (i + 1)] for i in range(int(1 / self.sample_size)-1)]
+
+            # adding last sample with all remaining features
+            feature_samples += [self.features[chunks * (int(1 / self.sample_size) - 1):]]
+
+            # iterating over each feature samples
+            best_features = []
+            for features in feature_samples:
+
+                # fitting association on features
+                best_features += self.fit_features(X, y, features, int(self.n_best // 2))
+        
+        # splitting in chunks not requested
+        else:
+            best_features = self.features[:]
+        
+        # final selection with all best_features selected
+        self.best_features = self.fit_features(X, y, best_features, self.n_best)
+            
+        # dropped features
+        self.dropped_features = [c for c in self.features if c not in self.best_features]
 
         return self
 
@@ -165,7 +233,7 @@ class FeatureSelector(BaseEstimator, TransformerMixin):
             Xc.copy()
 
         # filtering out unwanted features
-        Xc = X.drop([c for c in self.features if c not in self.best_features], axis=1)
+        Xc = Xc.drop(self.dropped_features, axis=1)
 
         return Xc
 
@@ -180,18 +248,19 @@ def nans_measure(active: bool, association: Dict[str, Any], x: Series, y: Series
       Maximum percentage of NaNs in a feature
     """
     
-    passed = True  # whether or not tests where passed
+    # whether or not tests where passed
+    if active:
     
-    nans = x.isnull()  # ckecking for nans
-    pct_nan = nans.mean()   # Computing percentage of nans
+        nans = x.isnull()  # ckecking for nans
+        pct_nan = nans.mean()   # Computing percentage of nans
     
-    # updating association
-    association.update({'pct_nan': pct_nan})
+        # updating association
+        association.update({'pct_nan': pct_nan})
     
-    # Excluding feature that have to many NaNs
-    passed = pct_nan < params.get('thresh_nan', 1.)
+        # Excluding feature that have to many NaNs
+        active = pct_nan < params.get('thresh_nan', 1.)
     
-    return passed, association
+    return active, association
 
 def dtype_measure(active: bool, association: Dict[str, Any], x: Series, y: Series=None, **params) -> Tuple[bool, Dict[str, Any]]:
     """ Gets dtype"""
@@ -210,18 +279,19 @@ def mode_measure(active: bool, association: Dict[str, Any], x: Series, y: Series
       Maximum percentage of the mode of a feature
     """
     
-    passed = True  # whether or not tests where passed
+    # whether or not tests where passed
+    if active:
     
-    mode = x.mode(dropna=True).values[0]  # computing mode
-    pct_mode = (x == mode).mean()  # Computing percentage of the mode
+        mode = x.mode(dropna=True).values[0]  # computing mode
+        pct_mode = (x == mode).mean()  # Computing percentage of the mode
     
-    # updating association
-    association.update({'pct_mode': pct_mode, 'mode': mode})
+        # updating association
+        association.update({'pct_mode': pct_mode, 'mode': mode})
     
-    # Excluding feature with too frequent modes
-    passed = pct_mode < params.get('thresh_mode', 1.)
+        # Excluding feature with too frequent modes
+        active = pct_mode < params.get('thresh_mode', 1.)
     
-    return passed, association
+    return active, association
 
 def kruskal_measure(active: bool, association: Dict[str, Any], x: Series, y: Series, **params) -> Tuple[bool, Dict[str, Any]]:
     """ Kruskal-Wallis statistic between x (quantitative) and y (binary)"""
@@ -256,7 +326,10 @@ def R_measure(active: bool, association: Dict[str, Any], x: Series, y: Series, *
         
         # updating association
         if regression.rsquared:
-            association.update({'R_measure': sqrt(regression.rsquared)})
+            if regression.rsquared >= 0:
+                association.update({'R_measure': sqrt(regression.rsquared)})
+            else:
+                association.update({'R_measure': nan})
         
     return active, association
 
@@ -291,7 +364,7 @@ def zscore_measure(active: bool, association: Dict[str, Any], x: Series, y: Seri
         })
 
         # Excluding feature with too frequent modes
-        passed = pct_zscore < params.get('thresh_outlier', 1.)
+        active = pct_zscore < params.get('thresh_outlier', 1.)
         
     return active, association
 
@@ -325,7 +398,7 @@ def iqr_measure(active: bool, association: Dict[str, Any], x: Series, y: Series=
         })
 
         # Excluding feature with too frequent modes
-        passed = pct_iqr < params.get('thresh_outlier', 1.)
+        active = pct_iqr < params.get('thresh_outlier', 1.)
         
     return active, association
 
@@ -414,7 +487,7 @@ def tschuprowt_measure(active: bool, association: Dict[str, Any],x: Series,
     return active, association
 
     
-# FILTERS        
+# FILTERS 
 def thresh_filter(X: DataFrame, ranks: DataFrame, **params) -> Dict[str, Any]:
     """ Filters out missing association measure (did not pass a threshold)"""
     
@@ -423,6 +496,27 @@ def thresh_filter(X: DataFrame, ranks: DataFrame, **params) -> Dict[str, Any]:
     
     return associations
 
+def measure_filter(X: DataFrame, ranks: DataFrame, **params) -> Dict[str, Any]:
+    """ Filters out specified measure's lower ranks than threshold
+
+    Parameters
+    ----------
+    thresh_measure, float: default 0.
+        Minimum association between target and features
+        To be used with: `association_filter`
+    name_measure, str
+        Measure to be used for minimum association filtering
+        To be used with: `association_filter`
+    """
+    
+    associations = ranks.copy()
+
+    # drops rows with nans
+    if 'name_measure' in params:
+        associations = ranks[ranks[params.get('name_measure')] > params.get('thresh_measure', 0.)]
+    
+    return associations
+
 def quantitative_filter(X: DataFrame, ranks: DataFrame, corr_measure: str, **params) -> Dict[str, Any]:
     """ Computes max association between X and X (quantitative) excluding features 
     that are correlated to a feature more associated with the target 
@@ -431,7 +525,7 @@ def quantitative_filter(X: DataFrame, ranks: DataFrame, corr_measure: str, **par
     Parameters
     ----------
     thresh_corr, float: default 1.
-      Maximum association between features
+        Maximum association between features
     """
     
     # accessing the prefered order
